import { resolve, join } from 'path'
import Vue from 'vue'
import {
  getCurrentInstance,
  onServerPrefetch,
  onBeforeMount,
} from '@vue/composition-api'
export * from '@vue/composition-api'
import { normalizeError } from '@nuxt/vue-app'

const fetches = new WeakMap()
const isSsrHydration = vm => {
  var _a, _b, _c
  return (_c =
    (_b = (_a = vm.$vnode) === null || _a === void 0 ? void 0 : _a.elm) ===
      null || _b === void 0
      ? void 0
      : _b.dataset) === null || _c === void 0
    ? void 0
    : _c.fetchKey
}
const nuxtState = process.client && window.__NUXT__
function registerCallback(vm, callback) {
  const callbacks = fetches.get(vm) || []
  fetches.set(vm, [...callbacks, callback])
}
async function callFetches() {
  const fetchesToCall = fetches.get(this)
  if (!fetchesToCall) return
  this.$nuxt.nbFetching++
  this.$fetchState =
    this.$fetchState ||
    Vue.observable({
      error: null,
      pending: false,
      timestamp: 0,
    })
  this.$fetchState.pending = true
  this.$fetchState.error = null
  this._hydrated = false
  let error = null
  const startTime = Date.now()
  try {
    await Promise.all(fetchesToCall.map(fetch => fetch(this)))
  } catch (err) {
    error = normalizeError(err)
  }
  const delayLeft = (this._fetchDelay || 0) - (Date.now() - startTime)
  if (delayLeft > 0) {
    await new Promise(resolve => setTimeout(resolve, delayLeft))
  }
  this.$fetchState.error = error
  this.$fetchState.pending = false
  this.$fetchState.timestamp = Date.now()
  this.$nextTick(() => this.$nuxt.nbFetching--)
}
async function serverPrefetch(vm) {
  // Call and await on $fetch
  vm.$fetchState =
    vm.$fetchState ||
    Vue.observable({
      error: null,
      pending: false,
      timestamp: 0,
    })
  try {
    await callFetches.call(vm)
  } catch (err) {
    vm.$fetchState.error = normalizeError(err)
  }
  vm.$fetchState.pending = false
  // Define an ssrKey for hydration
  vm._fetchKey = vm.$ssrContext.nuxt.fetch.length
  // Add data-fetch-key on parent element of Component
  if (!vm.$vnode.data) vm.$vnode.data = {}
  const attrs = (vm.$vnode.data.attrs = vm.$vnode.data.attrs || {})
  attrs['data-fetch-key'] = vm._fetchKey
  // Add to ssrContext for window.__NUXT__.fetch
  vm.$ssrContext.nuxt.fetch.push(
    vm.$fetchState.error
      ? { _error: vm.$fetchState.error }
      : JSON.parse(JSON.stringify(vm._data))
  )
}
const useFetch = callback => {
  var _a
  const vm = getCurrentInstance()
  if (!vm) throw new Error('This must be called within a setup function.')
  registerCallback(vm, callback)
  onServerPrefetch(serverPrefetch)
  onBeforeMount(() => {
    if (!vm._hydrated) {
      return callFetches.call(vm)
    }
  })
  if (process.server || !isSsrHydration(vm)) return
  // Hydrate component
  vm._hydrated = true
  vm._fetchKey = +((_a = vm.$vnode.elm) === null || _a === void 0
    ? void 0
    : _a.dataset.fetchKey)
  const data = nuxtState.fetch[vm._fetchKey]
  // If fetch error
  if (data && data._error) {
    vm.$fetchState.error = data._error
    return
  }
  onBeforeMount(() => {
    // Merge data
    for (const key in data) {
      try {
        Vue.set(vm, key, data[key])
      } catch (e) {
        if (process.env.NODE_ENV === 'development')
          // eslint-disable-next-line
          console.warn(`Could not hydrate ${key}.`)
      }
    }
  })
}

const withContext = callback => {
  const vm = getCurrentInstance()
  if (!vm) throw new Error('This must be called within a setup function.')
  callback(vm.$nuxt.context)
}

const compositionApiModule = function () {
  const libRoot = resolve(__dirname, '..')
  const { dst } = this.addTemplate({
    src: resolve(libRoot, 'lib', 'plugin.js'),
    fileName: join('composition-api', 'plugin.js'),
    options: {},
  })
  this.options.plugins = this.options.plugins || []
  this.options.plugins.push(resolve(this.options.buildDir || '', dst))
}
// eslint-disable-next-line
const meta = require('../package.json')

export default compositionApiModule
export { meta, useFetch, withContext }
